/*
 * Copyright (c) 2006-2007, 2009 by Freescale Semiconductor, Inc.  All rights reserved.
 *               2010 by Marc Kleine-Budde <kernel@pengutronix.de>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <btcs.h>
#include <btcs_config.h>
#include <candriver.h>
#include <canstack.h>

#include <linux/btcs_interface.h>

#define BTCS_WRITE_STRING(s) canstack_write_bytes(s, sizeof(s)-1)

static struct btcs_interface btcs_interface;
static struct btcs_shm btcs_shm;

extern char _stext;
extern char _edata;

static void btcs_prompt(void)
{
	BTCS_WRITE_STRING(">");
}

static void btcs_cr(void)
{
	BTCS_WRITE_STRING("\r\n");
}

static void btcs_ltohex(char *buf, unsigned long value)
{
	const char *hex = "0123456789abcdef";
	int i;

	for (i = 0; i < 8; i++)
		buf[7 - i] = hex[(value >> (4 * i)) & 0xf];
}

static int btcs_hextol(char *buf, unsigned long len)
{
	int val = 0;
	int i;
	for (i = 0; i < len; i++) {
		char c = buf[i];
		int digit;
		if ((c >= '0') && (c <= '9'))
			digit = c - '0';
		else if ((c >= 'a') && (c <= 'f'))
			digit = c - 'a' + 10;
		else if ((c >= 'A') && (c <= 'F'))
			digit = c - 'A' + 10;
		else
			return 0;

		val <<= 4;
		val |= digit;
	}
	return val;
}

static void btcs_showreg(int num)
{
	char buf[] = "[n] hhhhhhhh\r\n";
	buf[1] = num + '0';
	btcs_ltohex(&buf[4], btcs_shm.reg[num]);
	BTCS_WRITE_STRING(buf);
}

/* setup function */
static int btcs_setup(void)
{
	/*
	 * ld script puts bss in data so
	 * no need to zero it
	 */
	btcs_interface.size = &_edata - &_stext;
	canstack_init();
	candriver_init();
	return 0;
}

/* start function */
static int btcs_start(void)
{
	BTCS_WRITE_STRING("\r\n\r\n*** BTCS Demo Stack ***\r\n");
	btcs_prompt();
	candriver_enable_rx();
	return 0;
}

/* stop function */
static int btcs_stop(void)
{
	return 0;
}

/* cleanup function */
static int btcs_cleanup(void)
{
	return 0;
}

/* callback function */
static int btcs_callback(struct btcs_event *event)
{
	switch (event->cmd) {
	case BTCS_IRQ:
		candriver_irq();
		break;
	case BTCS_REG_MODIFIED:
		btcs_cr();
		btcs_showreg((int)event->arg);
		btcs_prompt();
		break;
	default:
		/* do nothing */
		break;
	}
	return 0;
}

static struct btcs_interface btcs_interface __attribute__((section(".btcs.text"))) = {
	.magic		= BTCS_MAGIC,
	.text_base	= CONFIG_BTCS_TEXT_BASE,
	.setup		= btcs_setup,
	.start		= btcs_start,
	.stop		= btcs_stop,
	.cleanup	= btcs_cleanup,
	.callback	= btcs_callback,
	.shm		= &btcs_shm,
};

int printk(void *fmt, ...)
{
	va_list args;
	int ret;

	if (!btcs_interface.vprintk)
		return 0;

	va_start(args, fmt);
	ret = btcs_interface.vprintk(fmt, args);
	va_end(args);

	return ret;
}

#define BTCS_CMD_LEN 78
static char btcs_cmd_string[BTCS_CMD_LEN];
static int btcs_cmd_len;

static void btcs_cmd_received(char *cmd, int len)
{
	if (len == 0) {
		/* do nothing */
	} else if ((len == 3) && (cmd[0] == 'd')) {
		int num = cmd[2] - '0';
		if ((num >= 0) && (num < BTCS_NUM_REGS))
			btcs_showreg(num);
		else
			BTCS_WRITE_STRING("!invalid argument\r\n");
	} else if ((len > 4) && (cmd[0] == 'm')) {
		unsigned int num = cmd[2] - '0';

		if ((num >= 0) && (num < BTCS_NUM_REGS)) {
			struct btcs_event event;
			unsigned int val = btcs_hextol(&cmd[4], len - 4);

			btcs_shm.reg[num] = val;
			if (btcs_interface.ossignal) {
				event.cmd = BTCS_REG_MODIFIED;
				event.arg = (void *)num;
				btcs_interface.ossignal(&event);
			}
		} else {
			BTCS_WRITE_STRING("!invalid argument\r\n");
		}
	} else if (cmd[0] == '#') {
		/* do nothing */
		;
	} else {
		BTCS_WRITE_STRING("!syntax error\r\n");
	}
	btcs_prompt();
}

static void btcs_char_received(char c)
{
	if (c == '\n')
		c = '\r';

	if ((btcs_cmd_len < BTCS_CMD_LEN) || (c == '\r') || (c == 0x7f)) {
		if (c == '\r') {
			btcs_cr();
			btcs_cmd_received(btcs_cmd_string, btcs_cmd_len);
			btcs_cmd_len = 0;
		} else if (c == 0x7f) {
			if (btcs_cmd_len > 0) {
				BTCS_WRITE_STRING("\b \b");
				btcs_cmd_len--;
			}
		} else {
			canstack_write_bytes(&c, 1);
			btcs_cmd_string[btcs_cmd_len] = c;
			btcs_cmd_len++;
		}
	}
}

void btcs_cb(void)
{
	int b;

	while (1) {
		b = canstack_read_byte();
		if (b < 0)
			break;

		btcs_char_received(b);
	}
	candriver_enable_tx();
}
