/*
 * Copyright (c) 2006-2007, 2009 Freescale Semiconductor, Inc. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/*
 * Simple CAN stack implementation.
 * Translates sequence of RX CAN packets to stream of input bytes
 * and stream of output bytes to sequence of TX CAN packets.
 */

#include <btcs.h>
#include <candriver.h>
#include <canstack.h>
#include <types.h>

#define RB_SIZE		(4 * 1024)
#define RB_MASK		(RB_SIZE - 1)
#define RB_BIGMASK	(2 * RB_SIZE - 1)

struct rb {
	u8 *buf;
	int rdindx;	/* rdindx & mask is read index */
	int wrindx;	/* wrindx & mask is write index */
};

/*
 * rdindx and wrindx range from 0 to RB_BIGMASK
 * 0 <= bytes available for read
 *   == (wrindx - rdindx) & RB_BIGMASK
 *   <= RB_SIZE
 *
 * 0 <= bytes available for write
 *   == RB_SIZE - bytes available for read
 *   <= RB_SIZE
 */

static int rb_rd_space(struct rb *rb)
{
	return (rb->wrindx - rb->rdindx) & RB_BIGMASK;
}

static int rb_wr_space(struct rb *rb)
{
	return RB_SIZE - rb_rd_space(rb);
}

static void rb_init(struct rb *rb, u8 *buf, int size)
{
	rb->buf = buf;
	rb->rdindx = 0;
	rb->wrindx = 0;
}

static void rb_rd_inc(struct rb *rb)
{
	rb->rdindx = (rb->rdindx + 1) & RB_BIGMASK;
}

static void rb_wr_inc(struct rb *rb)
{
	rb->wrindx = (rb->wrindx + 1) & RB_BIGMASK;
}

static void rb_write_byte(struct rb *rb, u8 b)
{
	rb->buf[rb->wrindx&RB_MASK] = b;
	rb_wr_inc(rb);
}

static int rb_write_bytes(struct rb *rb, u8 *buf, int len)
{
	int rv = len;

	if (rb_wr_space(rb) >= len) {
		while (len-- > 0)
			rb_write_byte(rb, *buf++);
	} else {
		rv = -1;
	}
	return rv;
}

static u8 rb_read_byte(struct rb *rb)
{
	u8 b = rb->buf[rb->rdindx&RB_MASK];

	rb_rd_inc(rb);
	return b;
}

static int rb_read_bytes(struct rb *rb, u8 *buf, int len)
{
	int rv;

	if (len >= rb_rd_space(rb))
		len = rb_rd_space(rb);
	rv = len;

	while (len-- > 0)
		*buf++ = rb_read_byte(rb);

	return rv;
}

static u8 rxbuf[RB_SIZE];
static struct rb rxrb;

static u8 txbuf[RB_SIZE];
static struct rb txrb;

void canstack_init(void)
{
	rb_init(&rxrb, rxbuf, RB_SIZE);
	rb_init(&txrb, txbuf, RB_SIZE);
}

int canstack_read_byte(void)
{
	int ret;

	ret = rb_rd_space(&rxrb);
	if (ret > 0)
		return rb_read_byte(&rxrb);
	else
		return -1;
}

int canstack_write_byte(u8 b)
{
	int ret;

	ret = rb_wr_space(&txrb);
	if (ret > 0) {
		rb_write_byte(&txrb, b);
		return 0;
	} else {
		return -1;
	}
}

int canstack_write_bytes(u8 *p, int len)
{
	while (len--)
		canstack_write_byte(*p++);

	candriver_enable_tx();

	return 0;
}

int canstack_txpending(void)
{
	int avail = rb_rd_space(&txrb);

	return avail > 0;
}

#define CAN_TX_ID	0x200
#define CAN_TX_FLAG	CAN_FLAG_DEFAULT

void canstack_dequeue_txpacket(struct canpacket *cp)
{
	int avail = rb_rd_space(&txrb);

	if (avail > 8)
		avail = 8;

	cp->id = CAN_TX_ID;
	cp->flags = CAN_TX_FLAG;
	cp->len = avail;
	rb_read_bytes(&txrb, cp->data, avail);
}

int canstack_queue_rxpacket(struct canpacket *cp)
{
	int rv;
	int avail = rb_wr_space(&rxrb);

	if (avail >= cp->len) {
		rb_write_bytes(&rxrb, cp->data, cp->len);
		rv = 1;
	} else {
		rv = 0;
	}

	btcs_cb();
	return rv;
}
