/*
 * Copyright (c) 2009 Freescale Semiconductor, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <candriver.h>
#include <canstack.h>
#include <io.h>
#include <misc.h>

struct can_mb_cs {
	unsigned int time_stamp:16;
	unsigned int length:4;
	unsigned int rtr:1;
	unsigned int ide:1;
	unsigned int srr:1;
	unsigned int nouse1:1;
	unsigned int code:4;
	unsigned int nouse2:4;
};

#define CAN_MB_RX_INACTIVE	0x0
#define CAN_MB_RX_EMPTY		0x4
#define CAN_MB_RX_FULL		0x2
#define CAN_MB_RX_OVERRUN	0x6
#define CAN_MB_RX_BUSY		0x1

#define CAN_MB_TX_INACTIVE	0x8
#define CAN_MB_TX_ABORT		0x9
#define CAN_MB_TX_ONCE		0xC
#define CAN_MB_TX_REMOTE	0xA

struct can_hw_mb {
	union {
		struct can_mb_cs cs;
		unsigned int data;
	} mb_cs;
	unsigned int mb_id;
	unsigned char mb_data[8];
};

#define CAN_HW_REG_MCR		0x00
#define CAN_HW_REG_CTRL		0x04
#define CAN_HW_REG_TIMER	0x08
#define CAN_HW_REG_RXGMASK	0x10
#define CAN_HW_REG_RX14MASK	0x14
#define CAN_HW_REG_RX15MASK	0x18
#define CAN_HW_REG_ECR		0x1C
#define CAN_HW_REG_ESR		0x20
#define CAN_HW_REG_IMASK2	0x24
#define CAN_HW_REG_IMASK1	0x28
#define CAN_HW_REG_IFLAG2	0x2C
#define CAN_HW_REG_IFLAG1	0x30

#define CAN_MB_BASE		0x0080
#define CAN_RXMASK_BASE		0x0880
#define CAN_FIFO_BASE		0xE0

#define __MCR_MDIS		(1 << 31)
#define __MCR_FRZ		(1 << 30)
#define __MCR_FEN		(1 << 29)
#define __MCR_HALT		(1 << 28)
#define __MCR_NOTRDY		(1 << 27)
#define __MCR_WAK_MSK		(1 << 26)
#define __MCR_SOFT_RST		(1 << 25)
#define __MCR_FRZ_ACK		(1 << 24)
#define __MCR_SLF_WAK		(1 << 22)
#define __MCR_WRN_EN		(1 << 21)
#define __MCR_LPM_ACK		(1 << 20)
#define __MCR_WAK_SRC		(1 << 19)
#define __MCR_DOZE		(1 << 18)
#define __MCR_SRX_DIS		(1 << 17)
#define __MCR_BCC		(1 << 16)
#define __MCR_LPRIO_EN		(1 << 13)
#define __MCR_AEN		(1 << 12)
#define __MCR_MAX_IDAM_OFFSET	8
#define __MCR_MAX_IDAM_MASK	(0x3 << __MCR_MAX_IDAM_OFFSET)
#define __MCR_MAX_IDAM_A	(0x0 << __MCR_MAX_IDAM_OFFSET)
#define __MCR_MAX_IDAM_B	(0x1 << __MCR_MAX_IDAM_OFFSET)
#define __MCR_MAX_IDAM_C	(0x2 << __MCR_MAX_IDAM_OFFSET)
#define __MCR_MAX_IDAM_D	(0x3 << __MCR_MAX_IDAM_OFFSET)
#define __MCR_MAX_MB_OFFSET	0
#define __MCR_MAX_MB_MASK	(0x3F)

#define __CTRL_PRESDIV_OFFSET	24
#define __CTRL_PRESDIV_MASK	(0xFF << __CTRL_PRESDIV_OFFSET)
#define __CTRL_RJW_OFFSET	22
#define __CTRL_RJW_MASK		(0x3 << __CTRL_RJW_OFFSET)
#define __CTRL_PSEG1_OFFSET	19
#define __CTRL_PSEG1_MASK	(0x7 << __CTRL_PSEG1_OFFSET)
#define __CTRL_PSEG2_OFFSET	16
#define __CTRL_PSEG2_MASK	(0x7 << __CTRL_PSEG2_OFFSET)
#define __CTRL_BOFF_MSK		(0x1 << 15)
#define __CTRL_ERR_MSK		(0x1 << 14)
#define __CTRL_CLK_SRC		(0x1 << 13)
#define __CTRL_LPB		(0x1 << 12)
#define __CTRL_TWRN_MSK		(0x1 << 11)
#define __CTRL_RWRN_MSK		(0x1 << 10)
#define __CTRL_SMP		(0x1 << 7)
#define __CTRL_BOFF_REC		(0x1 << 6)
#define __CTRL_TSYN		(0x1 << 5)
#define __CTRL_LBUF		(0x1 << 4)
#define __CTRL_LOM		(0x1 << 3)
#define __CTRL_PROPSEG_OFFSET	0
#define __CTRL_PROPSEG_MASK	(0x7)

#define __ECR_TX_ERR_COUNTER(x)	((x) & 0xFF)
#define __ECR_RX_ERR_COUNTER(x)	(((x) >> 8) & 0xFF)
#define __ECR_PASSIVE_THRESHOLD	128
#define __ECR_ACTIVE_THRESHOLD	96

#define __ESR_TWRN_INT		(0x1 << 17)
#define __ESR_RWRN_INT		(0x1 << 16)
#define __ESR_BIT1_ERR		(0x1 << 15)
#define __ESR_BIT0_ERR		(0x1 << 14)
#define __ESR_ACK_ERR		(0x1 << 13)
#define __ESR_CRC_ERR		(0x1 << 12)
#define __ESR_FRM_ERR		(0x1 << 11)
#define __ESR_STF_ERR		(0x1 << 10)
#define __ESR_TX_WRN		(0x1 << 9)
#define __ESR_RX_WRN		(0x1 << 8)
#define __ESR_IDLE		(0x1 << 7)
#define __ESR_TXRX		(0x1 << 6)
#define __ESR_FLT_CONF_OFF	4
#define __ESR_FLT_CONF_MASK	(0x3 << __ESR_FLT_CONF_OFF)
#define __ESR_BOFF_INT		(0x1 << 2)
#define __ESR_ERR_INT		(0x1 << 1)
#define __ESR_WAK_INT		(0x1)

#define __ESR_INTERRUPTS	(__ESR_WAK_INT | __ESR_ERR_INT |	\
				 __ESR_BOFF_INT | __ESR_TWRN_INT |	\
				 __ESR_RWRN_INT)

#define __FIFO_OV_INT		0x0080
#define __FIFO_WARN_INT		0x0040
#define __FIFO_RDY_INT		0x0020

struct flexcan_device {
	void *io_base;
	struct can_hw_mb *hwmb;
	unsigned int *rx_mask;
	unsigned int xmit_mb;
	unsigned int parent_clk_rate;
	/* word 1 */
	unsigned int br_presdiv:8;
	unsigned int br_rjw:2;
	unsigned int br_propseg:3;
	unsigned int br_pseg1:3;
	unsigned int br_pseg2:3;
	unsigned int maxmb:6;
	unsigned int xmit_maxmb:6;
	unsigned int wd1_resv:1;

	/* word 2 */
	unsigned int wakeup:1;
	unsigned int srx_dis:1;
	unsigned int wak_src:1;
	unsigned int bcc:1;
	unsigned int lprio:1;
	unsigned int abort:1;
	unsigned int br_clksrc:1;
	unsigned int loopback:1;
	unsigned int smp:1;
	unsigned int boff_rec:1;
	unsigned int tsyn:1;
	unsigned int listen:1;

	unsigned int ext_msg:1;
	unsigned int std_msg:1;
};

#define FLEXCAN_MAX_MB		64

#define AIPS2_BASE_ADDR_VIRT	0xFC200000
#define MX35_CAN2_BASE_ADDR	(AIPS2_BASE_ADDR_VIRT + 0xe8000)
#define FLEX_CAN_IOADDRESS	MX35_CAN2_BASE_ADDR

static void flexcan_device_attach(struct flexcan_device *flexcan)
{
	flexcan->io_base = (void *)FLEX_CAN_IOADDRESS;
	flexcan->hwmb = (struct can_hw_mb *)(flexcan->io_base + CAN_MB_BASE);
	flexcan->rx_mask = (unsigned int *)(flexcan->io_base + CAN_RXMASK_BASE);
}

static void flexcan_device_defaults(struct flexcan_device *dev)
{
	dev->br_clksrc = 1;

	/*
	 * Bit timing parameters for flexcan with 66.500000 MHz ref clock
	 * nominal                                 real Bitrt   nom  real SampP
	 * Bitrate TQ[ns] PrS PhS1 PhS2 SJW BRP Bitrate Error SampP SampP Error   CAN_CTRL
	 * 1000000     90   2    3    2   0   5 1007575  0.8% 75.0% 72.7%  3.1% 0x051a0002
	 *  800000    180   1    1    1   0  11  791666  1.0% 80.0% 71.4% 10.8% 0x0b090001
	 *  500000    105   6    7    2   0   6  500000  0.0% 87.5% 84.2%  3.8% 0x063a0006
	 *
	 *  250000    285   4    5    1   0  18  250000  0.0% 87.5% 85.7%  2.1% 0x12290004
	 *
	 *  125000    571   4    5    1   0  37  125000  0.0% 87.5% 85.7%  2.1% 0x25290004
	 *  100000    526   6    7    2   0  34  100000  0.0% 87.5% 84.2%  3.8% 0x223a0006
	 *   50000   1428   4    5    1   0  94   50000  0.0% 87.5% 85.7%  2.1% 0x5e290004
	 *   20000   2631   6    7    2   0 174   20000  0.0% 87.5% 84.2%  3.8% 0xae3a0006
	 *                  ^^^^^^^^^^^^^^^^^^^
	 * these are the actual register values.
	 *
	 */
	dev->br_propseg = 4;
	dev->br_pseg1 = 5;
	dev->br_pseg2 = 1;
	dev->br_rjw = 0;
	dev->br_presdiv = 18;

	dev->bcc = 1;
	dev->srx_dis = 1;
	dev->smp = 1;
	dev->boff_rec = 1;

	dev->maxmb = FLEXCAN_MAX_MB - 1;
	dev->xmit_maxmb = (FLEXCAN_MAX_MB >> 1) - 1;
	dev->xmit_mb = dev->maxmb - dev->xmit_maxmb;

	dev->ext_msg = 1;
	dev->std_msg = 1;
}

static struct flexcan_device __flexcan;

static struct flexcan_device *flexcan_device_alloc(void)
{
	struct flexcan_device *flexcan = &__flexcan;

	flexcan_device_attach(flexcan);
	flexcan_device_defaults(flexcan);

	return flexcan;
}

static void flexcan_hw_start(struct flexcan_device *flexcan)
{
	unsigned int reg;

	if ((flexcan->maxmb + 1) > 32) {
		__raw_writel(0xFFFFFFFF, flexcan->io_base + CAN_HW_REG_IMASK1);
		reg = (1 << (flexcan->maxmb - 31)) - 1;
		__raw_writel(reg, flexcan->io_base + CAN_HW_REG_IMASK2);
	} else {
		reg = (1 << (flexcan->maxmb + 1)) - 1;
		__raw_writel(reg, flexcan->io_base + CAN_HW_REG_IMASK1);
		__raw_writel(0, flexcan->io_base + CAN_HW_REG_IMASK2);
	}

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR) & (~__MCR_HALT);
	__raw_writel(reg, flexcan->io_base + CAN_HW_REG_MCR);
}

#if 0
static void flexcan_hw_stop(struct flexcan_device *flexcan)
{
	unsigned int reg;

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR);
	__raw_writel(reg | __MCR_HALT, flexcan->io_base + CAN_HW_REG_MCR);
}
#endif

static int flexcan_hw_reset(struct flexcan_device *flexcan)
{
	unsigned int reg;
	int timeout = 100000;

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR);
	__raw_writel(reg | __MCR_MDIS, flexcan->io_base + CAN_HW_REG_MCR);

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_CTRL);
	if (flexcan->br_clksrc)
		reg |= __CTRL_CLK_SRC;
	else
		reg &= ~__CTRL_CLK_SRC;
	__raw_writel(reg, flexcan->io_base + CAN_HW_REG_CTRL);

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR) & (~__MCR_MDIS);
	__raw_writel(reg, flexcan->io_base + CAN_HW_REG_MCR);
	reg |= __MCR_SOFT_RST;
	__raw_writel(reg, flexcan->io_base + CAN_HW_REG_MCR);

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR);
	while (reg & __MCR_SOFT_RST) {
		if (--timeout <= 0) {
			printk("Flexcan software reset timed out\n");
			return -1;
		}
		udelay(10);
		reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR);
	}
	return 0;
}

static inline void flexcan_mcr_setup(struct flexcan_device *flexcan)
{
	unsigned int reg;

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR);
	reg &= ~(__MCR_MAX_MB_MASK | __MCR_WAK_MSK | __MCR_MAX_IDAM_MASK);

	if (flexcan->wakeup)
		reg |= __MCR_SLF_WAK | __MCR_WAK_MSK;
	else
		reg &= ~(__MCR_SLF_WAK | __MCR_WAK_MSK);

	if (flexcan->wak_src)
		reg |= __MCR_WAK_SRC;
	else
		reg &= ~__MCR_WAK_SRC;

	if (flexcan->srx_dis)
		reg |= __MCR_SRX_DIS;
	else
		reg &= ~__MCR_SRX_DIS;

	if (flexcan->bcc)
		reg |= __MCR_BCC;
	else
		reg &= ~__MCR_BCC;

	if (flexcan->lprio)
		reg |= __MCR_LPRIO_EN;
	else
		reg &= ~__MCR_LPRIO_EN;

	if (flexcan->abort)
		reg |= __MCR_AEN;
	else
		reg &= ~__MCR_AEN;

	reg |= (flexcan->maxmb << __MCR_MAX_MB_OFFSET);
	__raw_writel(reg, flexcan->io_base + CAN_HW_REG_MCR);
}

static inline void flexcan_ctrl_setup(struct flexcan_device *flexcan)
{
	unsigned int reg;

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_CTRL);
	reg &= ~(__CTRL_PRESDIV_MASK | __CTRL_RJW_MASK | __CTRL_PSEG1_MASK |
		 __CTRL_PSEG2_MASK | __CTRL_PROPSEG_MASK);

	if (flexcan->loopback)
		reg |= __CTRL_LPB;
	else
		reg &= ~__CTRL_LPB;

	if (flexcan->smp)
		reg |= __CTRL_SMP;
	else
		reg &= ~__CTRL_SMP;

	if (flexcan->boff_rec)
		reg |= __CTRL_BOFF_REC;
	else
		reg &= ~__CTRL_BOFF_REC;

	if (flexcan->tsyn)
		reg |= __CTRL_TSYN;
	else
		reg &= ~__CTRL_TSYN;

	if (flexcan->listen)
		reg |= __CTRL_LOM;
	else
		reg &= ~__CTRL_LOM;

	reg |= (flexcan->br_presdiv << __CTRL_PRESDIV_OFFSET) |
		(flexcan->br_rjw << __CTRL_RJW_OFFSET) |
		(flexcan->br_pseg1 << __CTRL_PSEG1_OFFSET) |
		(flexcan->br_pseg2 << __CTRL_PSEG2_OFFSET) |
		(flexcan->br_propseg << __CTRL_PROPSEG_OFFSET);

	reg &= ~__CTRL_LBUF;

	reg |= __CTRL_TWRN_MSK | __CTRL_RWRN_MSK | __CTRL_BOFF_MSK |
		__CTRL_ERR_MSK;

	__raw_writel(reg, flexcan->io_base + CAN_HW_REG_CTRL);
}

static void flexcan_hw_init(struct flexcan_device *flexcan)
{
	if (flexcan_hw_reset(flexcan)) {
		printk("flexcan_hw_reset failed\n");
		return;
	}

	flexcan_mcr_setup(flexcan);
	flexcan_ctrl_setup(flexcan);

	__raw_writel(0, flexcan->io_base + CAN_HW_REG_IMASK2);
	__raw_writel(0, flexcan->io_base + CAN_HW_REG_IMASK1);

	__raw_writel(0xFFFFFFFF, flexcan->io_base + CAN_HW_REG_IFLAG2);
	__raw_writel(0xFFFFFFFF, flexcan->io_base + CAN_HW_REG_IFLAG1);

	__raw_writel(0, flexcan->io_base + CAN_HW_REG_ECR);
}

static void flexcan_err_handler(struct flexcan_device *flexcan)
{
	unsigned int esr;
	unsigned int reg;

	esr = __raw_readl(flexcan->io_base + CAN_HW_REG_ESR);

	__raw_writel(esr & __ESR_INTERRUPTS, flexcan->io_base + CAN_HW_REG_ESR);

	if (esr & __ESR_BOFF_INT) {
		reg = __raw_readl(flexcan->io_base + CAN_HW_REG_MCR);
		__raw_writel(reg | __MCR_SOFT_RST, flexcan->io_base + CAN_HW_REG_MCR);
	}
}

static unsigned int flexcan_swab32(unsigned int x)
{
	return (x & 0xff) << 24 |
		(x & 0xff00) << 8 |
		(x & 0xff0000) >> 8 |
		(x & 0xff000000) >> 24;
}

static inline void flexcan_memcpy(void *dst, void *src, int len)
{
	int i;
	unsigned int *d = (unsigned int *)dst, *s = (unsigned int *)src;
	len = (len + 3) >> 2;
	for (i = 0; i < len; i++)
		*d++ = flexcan_swab32(*s++);
}

static void flexcan_mb_bottom(struct flexcan_device *flexcan, int index)
{
	struct can_hw_mb *hwmb;
	struct canpacket zcp[1] = { { 0 } };
	struct canpacket *cp = zcp;

	hwmb = flexcan->hwmb + index;
	if (index >= (flexcan->maxmb - flexcan->xmit_maxmb)) {
		if (hwmb->mb_cs.cs.code == CAN_MB_TX_ABORT)
			hwmb->mb_cs.cs.code = CAN_MB_TX_INACTIVE;

		if (hwmb->mb_cs.cs.code & CAN_MB_TX_INACTIVE)
			return;
	}

	if (hwmb->mb_cs.cs.ide) {
		cp->id = hwmb->mb_id & CAN_PACKET_EXTD_MASK;
		cp->flags |= CAN_FLAG_EXTENDED;
	} else {
		cp->id = (hwmb->mb_id >> 18) & CAN_PACKET_STD_MASK;
	}

	if (hwmb->mb_cs.cs.rtr)
		cp->flags |= CAN_FLAG_RTR;

	cp->len = hwmb->mb_cs.cs.length;

	if (cp->len) {
		flexcan_memcpy(cp->data, hwmb->mb_data, cp->len);
		canstack_queue_rxpacket(cp);
	}

	if (index >= (flexcan->maxmb - flexcan->xmit_maxmb))
		hwmb->mb_cs.cs.code = CAN_MB_TX_INACTIVE;
}

static void flexcan_mbm_isr(struct flexcan_device *flexcan)
{
	int i, iflag1, iflag2, maxmb;

	if (flexcan->maxmb > 31) {
		maxmb = flexcan->maxmb + 1 - 32;
		iflag1 = __raw_readl(flexcan->io_base + CAN_HW_REG_IFLAG1) &
			__raw_readl(flexcan->io_base + CAN_HW_REG_IMASK1);
		iflag2 = __raw_readl(flexcan->io_base + CAN_HW_REG_IFLAG2) &
			__raw_readl(flexcan->io_base + CAN_HW_REG_IMASK2);
		iflag2 &= (1 << maxmb) - 1;
		maxmb = 32;
	} else {
		maxmb = flexcan->maxmb + 1;
		iflag1 = __raw_readl(flexcan->io_base + CAN_HW_REG_IFLAG1) &
			__raw_readl(flexcan->io_base + CAN_HW_REG_IMASK1);
		iflag1 &= (1 << maxmb) - 1;
		iflag2 = 0;
	}

	__raw_writel(iflag1, flexcan->io_base + CAN_HW_REG_IFLAG1);
	__raw_writel(iflag2, flexcan->io_base + CAN_HW_REG_IFLAG2);

	for (i = 0; iflag1 && (i < maxmb); i++) {
		if (iflag1 & (1 << i)) {
			iflag1 &= ~(1 << i);
			flexcan_mb_bottom(flexcan, i);
		}
	}

	for (i = maxmb; iflag2 && (i <= flexcan->maxmb); i++) {
		if (iflag2 & (1 << (i - 32))) {
			iflag2 &= ~(1 << (i - 32));
			flexcan_mb_bottom(flexcan, i);
		}
	}
}


static int flexcan_mbm_xmit(struct flexcan_device *flexcan, struct canpacket *cp)
{
	int i = flexcan->xmit_mb;
	struct can_hw_mb *hwmb = flexcan->hwmb;
	static int count;

	do {
		if (hwmb[i].mb_cs.cs.code == CAN_MB_TX_INACTIVE)
			break;
		if ((++i) > flexcan->maxmb)
			i = flexcan->xmit_maxmb + 1;
		if (i == flexcan->xmit_mb)
			return -1;
	} while (1);

	flexcan->xmit_mb = i + 1;
	if (flexcan->xmit_mb > flexcan->maxmb)
		flexcan->xmit_mb = flexcan->xmit_maxmb + 1;

	if (cp->flags & CAN_FLAG_RTR)
		hwmb[i].mb_cs.cs.rtr = 1;
	else
		hwmb[i].mb_cs.cs.rtr = 0;

	if (cp->flags & CAN_FLAG_EXTENDED) {
		hwmb[i].mb_cs.cs.ide = 1;
		hwmb[i].mb_cs.cs.srr = 1;
		hwmb[i].mb_id = cp->id;
	} else {
		hwmb[i].mb_cs.cs.ide = 0;
		hwmb[i].mb_id = cp->id << 18;
	}

	hwmb[i].mb_cs.cs.length = cp->len;
	flexcan_memcpy(hwmb[i].mb_data, cp->data, cp->len);
	hwmb[i].mb_cs.cs.code = CAN_MB_TX_ONCE;
	count++;

	return 0;
}

static void flexcan_mbm_init(struct flexcan_device *flexcan)
{
	struct can_hw_mb *hwmb;
	int rx_mb, i;

	/* Set global mask to receive all messages */
	__raw_writel(0, flexcan->io_base + CAN_HW_REG_RXGMASK);
	__raw_writel(0, flexcan->io_base + CAN_HW_REG_RX14MASK);
	__raw_writel(0, flexcan->io_base + CAN_HW_REG_RX15MASK);

	memset(flexcan->hwmb, 0, sizeof(*hwmb) * FLEXCAN_MAX_MB);
	/* Set individual mask to receive all messages */
	memset(flexcan->rx_mask, 0, sizeof(unsigned int) * FLEXCAN_MAX_MB);

	rx_mb = flexcan->maxmb - flexcan->xmit_maxmb;
	hwmb = flexcan->hwmb;
	for (i = 0; i < rx_mb; i++) {
		hwmb[i].mb_cs.cs.code = CAN_MB_RX_EMPTY;
		/*
		 * If we want to receive both extended and
		 * standard packets then only set the IDE bit
		 * in every other buffer.
		 * Buffers with ide set will get extended packets
		 * Buffers with ide clear will get standard packets
		 */
		if (flexcan->ext_msg && flexcan->std_msg)
			hwmb[i].mb_cs.cs.ide = i & 1;
		else {
			if (flexcan->ext_msg)
				hwmb[i].mb_cs.cs.ide = 1;
		}
	}

	for (; i <= flexcan->maxmb; i++)
		hwmb[i].mb_cs.cs.code = CAN_MB_TX_INACTIVE;

	flexcan->xmit_mb = rx_mb;
}

/*
 * Exported flexcan candriver interface
 */

/*
 * flexcan state used by most exported
 * candriver routines
 */
static struct flexcan_device *flexcan;

void candriver_init(void)
{
	flexcan = flexcan_device_alloc();

	flexcan_hw_init(flexcan);
	flexcan_mbm_init(flexcan);
	flexcan_hw_start(flexcan);
}

void candriver_enable_tx(void)
{
	struct canpacket cp;

	while (canstack_txpending()) {
		canstack_dequeue_txpacket(&cp);
		if (flexcan_mbm_xmit(flexcan, &cp) < 0)
			break;
	}
}

void candriver_irq(void)
{
	unsigned int reg;

	candriver_enable_tx();

	reg = __raw_readl(flexcan->io_base + CAN_HW_REG_ESR);
	if (reg & __ESR_INTERRUPTS) {
		flexcan_err_handler(flexcan);
		return;
	}

	flexcan_mbm_isr(flexcan);
	return;
}


void candriver_enable_rx(void)
{
}
